/*!  
  \file WeightedSupportCounterLookupSeq.hpp 
  \brief This is a weighted support counter (i.e. it implements SupportCounterInterface), 
  which takes the edges one-by-one and checks if there exists an element of 
  the transaction which is equal to the label of the edge.

  To find the corresponding transaction item, linear search is used.

  \author Ferenc BODON
  \date 2006-09-17
*/

#ifndef WeightedSupportCounterLookupSeq_HPP
#define WeightedSupportCounterLookupSeq_HPP



#include "common.hpp"
#include <vector>
#include "apriori/bodon/Leaf.hpp"
//#include <cstdio>
//#include <iterator>	//for test




namespace Bodon
{
namespace sequence
{
   template <class TRIE>
   class WeightedSupportCounterLookupSeq
   {
      protected:
	 typedef Leaf LEAF;
      public:
	 WeightedSupportCounterLookupSeq( ){}
	 void updateCounters(
	    TRIE& main_trie, const std::vector<item_t>& transaction,
	    item_t candidate_size, const counter_t counter_incr)
	 {
	    findCandidates( 
	       &main_trie, transaction.end()-candidate_size+1, 
	       transaction.begin(), candidate_size,
	       counter_incr );
	 }

      protected:
	 /// Increases the counter of those candidates that are contained 
	 /// by the given basket.
	 void findCandidates( TRIE* subtrie,
	    std::vector<item_t>::const_iterator it_transaction_upper_bound,
	    std::vector<item_t>::const_iterator it_transaction, 
	    item_t step_to_candidate,
	    const counter_t counter_incr );
   };

   template <class TRIE> void 
   WeightedSupportCounterLookupSeq<TRIE>::findCandidates( 
      TRIE* subtrie, std::vector<item_t>::const_iterator it_transaction_upper_bound,
      std::vector<item_t>::const_iterator it_transaction, 
      item_t step_to_candidate, const counter_t counter_incr )
   {
      std::vector<item_t>::const_iterator it_transaction2;
      --step_to_candidate;
      if( step_to_candidate ) 
      {
	 for( typename TRIE::iterator it_edge(subtrie->edgelist.begin()); 
	      it_edge != subtrie->edgelist.end(); ++it_edge)
	 {
	    for( it_transaction2 = it_transaction; 
		 it_transaction2 != it_transaction_upper_bound;
		 ++it_transaction2 )
	       if(*it_transaction2 == (*it_edge).first )
		  findCandidates( static_cast<TRIE*>((*it_edge).second),
				  it_transaction_upper_bound + 1, it_transaction2+1, 
				  step_to_candidate, counter_incr);
	 }
      }
      else
      {
	 for( typename TRIE::iterator it_edge(subtrie->edgelist.begin()); 
	      it_edge != subtrie->edgelist.end(); ++it_edge)
	 {
	    for( it_transaction2 = it_transaction; 
		 it_transaction2 != it_transaction_upper_bound;
		 ++it_transaction2 )
	       if(*it_transaction2 == (*it_edge).first )
		  static_cast<LEAF*>((*it_edge).second)->increaseCounter( counter_incr);
	 }
      }
   }
}
}
#endif
