/*!
  \file   main-fsm.cpp
  \brief  The main function of frequent itemsequence mining algorithms.

  Mining frequent sequence of item is a natural generalization of
  frequent itemset mining (FIM). Given a sequence of transactions,
  where each transaction is a sequence over an alphabet \f$I\f$, we
  have to find the sequences that occure as a subsequence in at least
  \f$min\_supp\f$ number of transactions. A sequence \f$s=\langle i_1,
  i_2, \ldots, i_n\rangle\f$ is a subsequence of \f$s'=\langle i'_1,
  i'_2, \ldots, i'_m\rangle\f$ if there exist integers \f$1\le j_1 <
  j_2 < \cdots < j_n \le m\f$, such that \f$i_1 = i'_{j_1}, i_2 =
  i'_{j_2}, \ldots, i_n = i'_{j_n}\f$, i.e. we can get $s$ by deleting
  some items from \f$s'\f$. For example \f$\langle e, a, a, b\rangle
  \prec \langle f, e, a, b, c, a, a, c, b\rangle\f$ because \f$i_1 =
  2$, $i_2 = 3$, $i_4 = 6$, $i_4 = 9\f$ meet the requirements.
  
  Currently this program only contains a trie-based Apriori
  implementation. For more information about this solution the reader
  is referred to the paper "A Trie-based APRIORI Implementation for
  Mining Frequent Item sequences" from Ferenc Bodon. It can be
  downloaded from the webpage of OSDM'05 workshop of ACM SIGKDD 
  (http://www.cs.rpi.edu/~zaki/OSDM05/papers/p56-bodon.pdf).

  \author Ferenc BODON
  \date 2005-04-19
*/

#include "common.hpp"
#include "io/input/transaction_reader/brBufferedTransactionReader.hpp"

#include "io/codec/decoder/df/CacheDFDecoder.hpp"
//#include "io/codec/decoder/df/DFDecoderWrapper.hpp"

#include "util/StreamParser.hpp"
#include "util/SeqFrequentFilter.cpp"

#include "datastructures/maxvector.hpp"
#include "datastructures/trie/edgelist/OrderedEdgelist.hpp"
//#include "datastructures/trie/edgelist/OffsetIndexVector.hpp"

#include "apriori/bodon/Trie.hpp"

#include "apriori/SeqAprioriSelector.hpp"
#include <vector>
#include <iostream>
#include <string>


std::string file_format;

void init()
{
   file_format = "File format:";
   file_format += "\n\nThe transactionfile is a plan text file. Each row ";
   file_format += "represents a transaction. \n";
   file_format += "A transaction is a set of items seperated by a nonnumeric ";
   file_format += "character.\nIt can be for example a white space, comma, ";
   file_format += "colon, etc.\n";
   file_format += "Items are nonnegative integers.\n";
}
/// This procedure displays the usage of the program.
void usage()
{
   std::cerr << "\nUsage: fsm algorithm transactionfile min_supp ";
   std::cerr << "outcomefile [maxsize]\n";
   std::cerr << "\n algorithm\t    the name of the algorithm, i.e:\n";
   std::cerr << "\t\t    apriori, apriori-noprune, apriori-intersectprune";
   std::cerr << "\n transactionfile    file, that contains the tranasctions of items";
   std::cerr << "\n outcomefile\t    file to write the outcome";
   std::cerr << "\n min_supp\t    absolute support threshold";
   std::cerr << "\n maxsize\t    the upper limit of the size of the frequent sets\n";

   std::cerr << file_format;
   std::cerr << "\n\nHave a succesful mining ;-)"<<std::endl<<std::endl;
}

/** This procedure processes the arguments.
   \return 
   <ul>
   <li>0, if no error is occured, </li>
   <li>1, in case of an IO error, </li>
   <li>2, in case of too few arguments,  </li>
   <li>3, if proper min_supp cannot be generated.  </li>
   </ul>
 */

int process_arguments( int argc, char *argv[], counter_t& min_supp, 
		       bool &isrel, double &relminsupp, unsigned int& maxsize )
{
   if ( argc < 5 )
   {
     log_err(0,"There are 4 mandatory arguments.");
     usage();
     return 2;
   }
   std::string mins=argv[3];
   if (mins[mins.size()-1]=='%') {
     mins.erase(mins.size()-1);
     isrel=true;
     relminsupp=atof(mins.c_str());
     relminsupp/=100;
     log_status(0,"Using relative minimum support of %lg",relminsupp);
     return 0;
   }
   isrel=false;
   int min_supp_i;
   try
   {
      convert(argv[3], min_supp_i);
      if ( min_supp_i <= 0  )
      {
	 log_err(0,"%s cannot be converted to a positive integer.",argv[3]);
	 return 3;
      }
   }
   catch(BadConversion e)
   {
      log_err(0,"min_supp conversion problem.");
      return 3;
   }
   min_supp = static_cast<counter_t>(min_supp_i);
   log_status(0,"min_supp is set to %d", min_supp);
   if(argc == 6)
   {
      int maxsize_i;
      try
      {
	 convert(argv[5], maxsize_i);
	 if ( maxsize_i <= 0  )
	 {
	    log_err(0,"%s cannot be converted to a positive integer.",argv[5]);
	    return 4;
	 }
      }
      catch(BadConversion e)
      {
	 log_err(0,"max_size conversion problem.");
	 return 4;
      }
      maxsize = static_cast<unsigned int>(maxsize_i);
      log_status(0,"maxsize is set to %d", maxsize);
   }
   else
      maxsize = largest_itemsetsize;
   return 0;
}

int main( int argc, char *argv[] )
{
   init();
   counter_t min_supp;
   unsigned int maxsize;
   bool relative;
   double relminsupp;
   
   {
      int return_val = 
	 process_arguments( argc, argv, min_supp, relative, relminsupp, maxsize );
      if(return_val)
	 return return_val;
   }

   if( strncmp(argv[1],"apriori",7) == 0 )
   {

      char* algorithm = argv[1];
      char* input_file = argv[2];
      char* output_file = argv[4];

      try
      {
	 typedef brBufferedTransactionReader< > T_R;
	 T_R::params_t par_i;
	 par_i.file_name = input_file;
	 par_i.mode=FileReprBase::READ;
	 par_i.file_buffer_size = 16 * 1024;
	 T_R tr_reader(&par_i);
	 std::vector< std::pair<counter_t, item_t> > freq_items_with_counters;
	 counter_t nr_of_transactions;
	 // The first step of each algorithms is determining the frequent items.
	 SeqFrequentFilter<T_R>
	    fr_filter(tr_reader);
	 log_status(0,"Finding frequent items.");
	 fr_filter.findFrequentItems( freq_items_with_counters,  
				      nr_of_transactions, min_supp);

	 log_status(0,"Doing decoder.");
	 typedef CacheDFDecoder<  > DF_D;
	 // typedef StatOutput< CacheDFDecoder< > >DF_D;

	 DF_D::params_t par_d;
	 par_d.file_name = output_file;
	 par_d.mode=FileReprBase::WRITE;
//	 par_d.numfreq = freq_items_with_counters.size();
	 DF_D df_decoder(&par_d);
	 typedef Bodon::LeafWithoutConstructor LEAF_WC;	 
	 typedef Bodon::Leaf LEAF;	 
	 if(strncmp(argv[1],"apriori",7) == 0)
	 {
	    log_status(0,"APRIORI is selected");
//	typedef Bodon::Trie< Bodon::OrderedEdgelist<maxvector<Edge> > > TRIE_BASE;
	    typedef Bodon::Trie< LEAF, Bodon::OrderedEdgelist<std::vector<Edge> > > TRIE_BASE;
//      typedef Bodon::Trie< Bodon::OffsetIndexVector< std::vector<void*> > > TRIE_BASE;
	 
	    SeqAprioriSelector<TRIE_BASE, LEAF_WC, T_R, DF_D>( 
	       min_supp, algorithm, input_file, nr_of_transactions, 
	       freq_items_with_counters, tr_reader, df_decoder, maxsize );
	 }
      }
         catch (std::ios_base::failure e)
	 {
	    log_err(0,"Exiting the program due to IO exception");
	    return 1;
	 }
   }
   else
   {
      usage();
      log_err(0,"algorithm should be apriori");
      return 1;
   }
}


